# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

from collections.abc import Iterable
from typing import TypeVar, Union

import numpy as np

from ._utility import get_rank

ellipsis = TypeVar("ellipsis")
ScalarIndexType = Union[int, bool, slice, ellipsis, None]


def index_normalise(a: Iterable[ScalarIndexType]) -> tuple[ScalarIndexType, ...]:
    ret: list[ScalarIndexType] = []
    for x in a:
        # x: Union[int, bool, slice, None]
        if isinstance(x, (int, bool)):
            ret.append(x)
        elif isinstance(x, slice):
            INDEX_MIN: int = np.iinfo(np.int64).min
            INDEX_MAX: int = np.iinfo(np.int64).max
            start = (
                x.start
                if x.start is not None
                else (0 if x.step is None or x.step > 0 else INDEX_MAX)
            )
            stop = (
                x.stop
                if x.stop is not None
                else (INDEX_MAX if x.step is None or x.step > 0 else INDEX_MIN)
            )
            step = x.step if x.step is not None else 1

            if start == 0 and stop == INDEX_MAX and step == 1:
                ret.append(slice(None, None, None))
            else:
                ret.append(slice(start, stop, step))
        elif x is None:
            ret.append(x)
        else:
            # ellipses are expected to be handled prior to this function call
            raise TypeError(f"Index {x} for type {type(x)} not supported")

    return tuple(ret)


def construct_index(
    arr, index: ScalarIndexType | tuple[ScalarIndexType, ...]
) -> tuple[ScalarIndexType, ...]:
    index_ = index if isinstance(index, tuple) else (index,)
    if any(i is Ellipsis for i in index_):
        rank = get_rank(arr)
        ellipsis_position = index_.index(Ellipsis)
        count_some = len(
            [x for x in index_ if not isinstance(x, (type(None), type(Ellipsis)))]
        )
        index_ = (
            index_[:ellipsis_position]
            + tuple([slice(None, None, None)] * (rank - count_some))
            + index_[ellipsis_position + 1 :]
        )
    return index_normalise(index_)



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

import math
import warnings

import numpy as np
import pytest

import ndonnx as ndx
import ndonnx.additional as nda

from .utils import assert_array_equal, get_numpy_array_api_namespace, run


def testfill_null():
    a = ndx.array(shape=(3,), dtype=ndx.nint64)
    b = nda.fill_null(a, ndx.asarray(0))

    model = ndx.build({"a": a}, {"b": b})

    expected_b = np.array([1, 2, 0], dtype=np.int64)
    assert_array_equal(
        run(
            model, {"a": np.ma.masked_array([1, 2, -1], mask=[0, 0, 1], dtype=np.int64)}
        )["b"],
        expected_b,
    )


@pytest.mark.parametrize(
    "fn_name",
    ["add", "subtract", "multiply", "divide", "floor_divide", "remainder"],
)
def test_arithmetic_none_propagation(fn_name):
    fn = getattr(ndx, fn_name)
    np_fn = getattr(np.ma, fn_name)

    a = ndx.array(shape=(3,), dtype=ndx.nfloat64)
    b = ndx.array(shape=(3,), dtype=ndx.nfloat64)
    c = fn(a, b)

    a_val = np.ma.masked_array([-1, 2.0, 3.0], mask=[1, 0, 0])
    b_val = np.ma.masked_array([2.0, 1.0, -1], mask=[0, 0, 1])

    model = ndx.build({"a": a, "b": b}, {"c": c})

    ret_c = run(model, {"a": a_val, "b": b_val})["c"]
    expected_c = np_fn(a_val, b_val)

    assert_array_equal(ret_c, expected_c)


@pytest.mark.parametrize(
    "fn_name, default_value",
    [
        ("sum", 0),
        # ("prod", 1),
        ("min", math.inf),
        ("max", -math.inf),
    ],
)
def test_reduce_ops_none_filling(fn_name, default_value):
    fn = getattr(ndx, fn_name)
    np_fn = getattr(np, fn_name)

    a = ndx.array(shape=(3,), dtype=ndx.nfloat32)
    b = fn(a)

    model = ndx.build({"a": a}, {"b": b})

    inp_a = np.ma.masked_array([0.0, -2.0, 3.0], dtype=np.float32, mask=[1, 0, 0])

    ret_b = run(model, {"a": inp_a})["b"]
    expected_b = np_fn(np.ma.filled(inp_a, default_value))

    assert_array_equal(ret_b, expected_b)


@pytest.mark.parametrize(
    "fn_name, args, kwargs",
    [
        ("abs", (), {}),
        ("exp", (), {}),
        ("log", (), {}),
        ("sqrt", (), {}),
        ("sin", (), {}),
        ("cos", (), {}),
        ("tan", (), {}),
        ("asin", (), {}),
        ("acos", (), {}),
        ("atan", (), {}),
        # ("sinh", (), {}),
        ("negative", (), {}),
        ("positive", (), {}),
        ("floor", (), {}),
        ("sign", (), {}),
        ("expand_dims", (), {"axis": 0}),
        ("reshape", ((3, 1),), {}),
        ("flip", (), {}),
        ("permute_dims", ((1, 0),), {}),
        ("roll", (1,), {}),
        ("squeeze", (0,), {}),
    ],
)
def test_unary_none_propagation(fn_name, args, kwargs):
    fn = getattr(ndx, fn_name)

    a = ndx.array(shape=(1, 3), dtype=ndx.nfloat32)
    b = fn(a, *args)

    model = ndx.build({"a": a}, {"b": b})

    inp_a = np.ma.masked_array([[0, -2.0, 3.0]], mask=[[1, 0, 0]], dtype=np.float32)
    ret_b = run(model, {"a": inp_a})["b"]
    missing_a = inp_a.mask
    npx = get_numpy_array_api_namespace()
    np_fn = getattr(npx, fn_name)
    inp_a = npx.asarray(np.ma.filled(inp_a, np.nan))

    # Numpy might complain about invalid values
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        expected_b = np_fn(inp_a, *args, **kwargs)
    np.testing.assert_almost_equal(
        np.ma.masked_array(expected_b, mask=missing_a),
        ret_b,
        decimal=5,
    )


def test_forbidden_masked():
    a = ndx.array(shape=(3,), dtype=ndx.nint64)

    with pytest.raises(TypeError):
        ndx.arange(a, 0, 1)


def test_masked_getitem():
    a = ndx.array(shape=(3,), dtype=ndx.nint64)

    model = ndx.build({"a": a}, {"b": a[0]})

    np.testing.assert_equal(
        [1],
        run(
            model, {"a": np.ma.masked_array([1, 2, 3], mask=[0, 0, 1], dtype=np.int64)}
        )["b"],
    )


def test_masked_setitem():
    a_start = ndx.array(shape=(3,), dtype=ndx.nint64)

    a = a_start.copy()
    a[0] = 1

    model = ndx.build({"a": a_start}, {"b": a})

    np.testing.assert_equal(
        [1, 2, 3],
        run(
            model, dict(a=np.ma.masked_array([1, 2, 3], mask=[0, 0, 0], dtype=np.int64))
        )["b"],
    )


def test_asarray_masked():
    a = ndx.asarray(np.ma.masked_array([1, 2, 3], mask=[0, 0, 1])) + 1
    assert_array_equal(
        a.to_numpy(),
        np.ma.masked_array([2, 3, 4], mask=[0, 0, 1]),
    )


def test_opset_extensions():
    import ndonnx._opset_extensions as opx
    from ndonnx._corearray import _CoreArray

    a = ndx.asarray(
        np.array([1, 2, 3]),
        dtype=ndx.int64,
    )
    shape = opx.shape(a.data)  # type: ignore
    # opset_extensions is an internal package that only deals with the internal state (lazy Spox Var and any eager values)
    assert isinstance(shape, _CoreArray)
    assert_array_equal(shape.to_numpy(), np.array([3], dtype=np.int64))


def test_eager_mode():
    a = ndx.asarray(np.ma.masked_array([1, 2, 3], mask=[0, 0, 1], dtype=np.int64))
    b = ndx.asarray(np.ma.masked_array([1, 2, 3], mask=[0, 0, 1], dtype=np.int64))
    c = ndx.asarray(
        np.ma.masked_array([-12, 21, 12213], mask=[1, 0, 0], dtype=np.int64)
    )
    assert_array_equal(
        (a + b).to_numpy(),
        np.ma.masked_array([2, 4, 6], mask=[0, 0, 1], dtype=np.int64),
    )
    assert_array_equal(
        (a - b).to_numpy(),
        np.ma.masked_array([0, 0, 0], mask=[0, 0, 1], dtype=np.int64),
    )
    assert_array_equal(
        (a * c).to_numpy(),
        np.ma.masked_array([-12, 42, 36639], mask=[1, 0, 1], dtype=np.int64),
    )


@pytest.mark.parametrize(
    "func",
    [
        np.tril,
        np.triu,
    ],
)
def test_trilu_masked_input(func):
    a = np.ma.masked_array(
        [[1, 2, 3], [4, 5, 6]], mask=[[0, 0, 1], [0, 0, 0]], dtype=np.int64
    )
    expected = func(a)
    actual = getattr(ndx, func.__name__)(ndx.asarray(a))
    assert_array_equal(actual.to_numpy(), expected)


@pytest.mark.parametrize(
    "arrays",
    [
        [
            np.ma.masked_array([1, 2, 3], mask=[0, 0, 1], dtype=np.int64),
            np.ma.masked_array([1, 2, 3], mask=[0, 0, 1], dtype=np.int64),
        ],
        [],
        [np.ma.masked_array([1, 2, 3], mask=[0, 0, 1], dtype=np.int64)],
        [
            np.ma.masked_array([1, 2, 3], mask=[0, 0, 1], dtype=np.int64),
            np.ma.masked_array(["a", "bc", "d"], mask=[0, 0, 1], dtype=np.str_),
        ],
        [
            np.ma.masked_array([1, 2, 3], mask=[0, 0, 1], dtype=np.int64),
            np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float64),
            np.array("a"),
        ],
        [
            np.array([1, 2, 3], dtype=np.int64),
            np.ma.masked_array(["a", "bc", "d"], mask=[0, 0, 1], dtype=np.str_),
        ],
    ],
)
def test_broadcasting(arrays):
    expected = np.broadcast_arrays(*arrays)
    actual = ndx.broadcast_arrays(*[ndx.asarray(a) for a in arrays])
    for e, a in zip(expected, actual):
        # NumPy simply drops the masked array.
        # We do not want to do the same quite intentionally.
        np.testing.assert_equal(a.to_numpy(), e)


@pytest.mark.parametrize(
    "np_array",
    [
        np.ma.masked_array([1, 2, 3], mask=[0, 0, 1], dtype=np.int64),
        np.ma.masked_array([1, 2, 3], mask=[0, 0, 1], dtype=np.float64),
        np.ma.masked_array([1, 2]),
        np.ma.masked_array(["a", "b"], mask=True),
        np.ma.masked_array([1, 2, 3], mask=[[[0]]]),
        np.ma.masked_array([[1, 2, 3]], mask=[True, False, True]),
        np.ma.masked_array([1.0, 2.0, 3.0], mask=[0, 0, 1]),
    ],
)
def test_initialization(np_array):
    actual = ndx.asarray(np_array)
    values = actual.values.to_numpy()
    null = actual.null.to_numpy()
    assert_array_equal(actual.to_numpy(), np_array)
    assert values.shape == null.shape



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import builtins
import typing
from collections.abc import Mapping
from typing import TypeVar

import numpy as np
import spox.opset.ai.onnx.ml.v4 as ml
import spox.opset.ai.onnx.v20 as op
from spox import Var
from spox._internal_op import unsafe_reshape

import ndonnx._data_types as dtypes

from ._corearray import _CoreArray
from ._index import ScalarIndexType
from ._propagation import eager_propagate
from ._utility import get_dtype, get_rank, get_shape

if typing.TYPE_CHECKING:
    from ndonnx._data_types.coretype import CoreType

ellipsis = TypeVar("ellipsis")


KeyType = TypeVar("KeyType", int, str, float)
ValueType = TypeVar("ValueType", int, str, float)


@eager_propagate
def add(x: _CoreArray, y: _CoreArray) -> _CoreArray:
    return _CoreArray(
        op.add(x.var, y.var),
    )


@eager_propagate
def string_concat(x: _CoreArray, y: _CoreArray) -> _CoreArray:
    return _CoreArray(
        op.string_concat(x.var, y.var),
    )


@eager_propagate
def expand(input: _CoreArray, shape: _CoreArray) -> _CoreArray:
    return _CoreArray(
        op.expand(input.var, shape.var),
    )


@eager_propagate
def asin(x: _CoreArray) -> _CoreArray:
    return _CoreArray(op.asin(x.var))


@eager_propagate
def range(start: _CoreArray, limit: _CoreArray, delta: _CoreArray) -> _CoreArray:
    return _CoreArray(op.range(start.var, limit.var, delta.var))


@eager_propagate
def asinh(x: _CoreArray) -> _CoreArray:
    return _CoreArray(op.asinh(x.var))


@eager_propagate
def atan(x: _CoreArray) -> _CoreArray:
    return _CoreArray(op.atan(x.var))


@eager_propagate
def atanh(x: _CoreArray) -> _CoreArray:
    return _CoreArray(op.atanh(x.var))


@eager_propagate
def or_(x: _CoreArray, y: _CoreArray) -> _CoreArray:
    return _CoreArray(op.or_(x.var, y.var))


@eager_propagate
def mul(x: _CoreArray, y: _CoreArray) -> _CoreArray:
    return _CoreArray(op.mul(x.var, y.var))


@eager_propagate
def cast(x: _CoreArray, to: CoreType) -> _CoreArray:
    return _CoreArray(
        op.cast(x.var, to=to.to_numpy_dtype()),
    )


@eager_propagate
def shape(x: _CoreArray, *, end: int | None = None, start: int = 0) -> _CoreArray:
    return _CoreArray(op.shape(x.var, end=end, start=start))


@eager_propagate
def top_k(
    x: _CoreArray, k: _CoreArray, *, axis: int = -1, largest: int = 1, sorted: int = 1
) -> tuple[_CoreArray, _CoreArray]:
    values, indices = op.top_k(x.var, k.var, axis=axis, largest=largest, sorted=sorted)
    return _CoreArray(values), _CoreArray(indices)


@eager_propagate
def where(cond: _CoreArray, x: _CoreArray, y: _CoreArray) -> _CoreArray:
    return _CoreArray(op.where(cond.var, x.var, y.var))


@eager_propagate
def not_(x: _CoreArray) -> _CoreArray:
    return _CoreArray(op.not_(x.var))


@eager_propagate
def eye_like(
    input: _CoreArray, *, dtype: CoreType | None = None, k: int = 0
) -> _CoreArray:
    return _CoreArray(
        op.eye_like(
            input.var, dtype=dtype if dtype is None else dtype.to_numpy_dtype(), k=k
        )
    )


@eager_propagate
def gather_elements(
    data: _CoreArray, indices: _CoreArray, *, axis: int = 0
) -> _CoreArray:
    return _CoreArray(op.gather_elements(data.var, indices.var, axis=axis))


@eager_propagate
def equal(x: _CoreArray, y: _CoreArray) -> _CoreArray:
    return _CoreArray(op.equal(x.var, y.var))


@eager_propagate
def reshape(x: _CoreArray, shape: _CoreArray) -> _CoreArray:
    return _CoreArray(op.reshape(x.var, shape.var))


@eager_propagate
def mod(x: _CoreArray, y: _CoreArray, fmod: int = 0) -> _CoreArray:
    return _CoreArray(op.mod(x.var, y.var, fmod=fmod))


@eager_propagate
def pow(x: _CoreArray, y: _CoreArray) -> _CoreArray:
    return _CoreArray(op.pow(x.var, y.var))


@eager_propagate
def div(x: _CoreArray, y: _CoreArray) -> _CoreArray:
    return _CoreArray(op.div(x.var, y.var))


@eager_propagate
def sub(x: _CoreArray, y: _CoreArray) -> _CoreArray:
    return _CoreArray(op.sub(x.var, y.var))


@eager_propagate
def ceil(x: _CoreArray) -> _CoreArray:
    return _CoreArray(op.ceil(x.var))


@eager_propagate
def transpose(x: _CoreArray, perm: list[int]) -> _CoreArray:
    return _CoreArray(op.transpose(x.var, perm=perm))


@eager_propagate
def const(value, dtype: CoreType | None = None) -> _CoreArray:
    np_dtype = dtype.to_numpy_dtype() if dtype is not None else None
    return _CoreArray(
        np.array(value, dtype=np_dtype),
    )


@eager_propagate
def squeeze(data: _CoreArray, axes: _CoreArray | None = None) -> _CoreArray:
    if axes is not None and axes.dtype != dtypes.int64:
        raise ValueError(f"Expected axes to be of type int64, got {axes.dtype}")
    return _CoreArray(op.squeeze(data.var, axes=axes.var if axes is not None else None))


@eager_propagate
def abs(x: _CoreArray) -> _CoreArray:
    return _CoreArray(op.abs(x.var))


@eager_propagate
def sin(x: _CoreArray) -> _CoreArray:
    return _CoreArray(op.sin(x.var))


@eager_propagate
def cos(x: _CoreArray) -> _CoreArray:
    return _CoreArray(op.cos(x.var))


@eager_propagate
def log(x: _CoreArray) -> _CoreArray:
    return _CoreArray(op.log(x.var))


@eager_propagate
def exp(x: _CoreArray) -> _CoreArray:
    return _CoreArray(op.exp(x.var))


@eager_propagate
def tan(x: _CoreArray) -> _CoreArray:
    return _CoreArray(op.tan(x.var))


@eager_propagate
def acos(x: _CoreArray) -> _CoreArray:
    return _CoreArray(op.acos(x.var))


@eager_propagate
def neg(x: _CoreArray) -> _CoreArray:
    return _CoreArray(op.neg(x.var))


@eager_propagate
def sqrt(x: _CoreArray) -> _CoreArray:
    return _CoreArray(op.sqrt(x.var))


@eager_propagate
def floor(x: _CoreArray) -> _CoreArray:
    return _CoreArray(op.floor(x.var))


@eager_propagate
def tanh(x: _CoreArray) -> _CoreArray:
    return _CoreArray(op.tanh(x.var))


@eager_propagate
def sinh(x: _CoreArray) -> _CoreArray:
    return _CoreArray(op.sinh(x.var))


@eager_propagate
def cosh(x: _CoreArray) -> _CoreArray:
    return _CoreArray(op.cosh(x.var))


@eager_propagate
def round(x: _CoreArray) -> _CoreArray:
    return _CoreArray(op.round(x.var))


@eager_propagate
def less(x: _CoreArray, y: _CoreArray) -> _CoreArray:
    return _CoreArray(op.less(x.var, y.var))


@eager_propagate
def less_or_equal(x: _CoreArray, y: _CoreArray) -> _CoreArray:
    return _CoreArray(op.less_or_equal(x.var, y.var))


@eager_propagate
def greater(x: _CoreArray, y: _CoreArray) -> _CoreArray:
    return _CoreArray(op.greater(x.var, y.var))


@eager_propagate
def greater_or_equal(x: _CoreArray, y: _CoreArray) -> _CoreArray:
    return _CoreArray(op.greater_or_equal(x.var, y.var))


@eager_propagate
def isnan(x: _CoreArray) -> _CoreArray:
    return _CoreArray(op.isnan(x.var))


@eager_propagate
def isinf(x: _CoreArray) -> _CoreArray:
    return _CoreArray(op.isinf(x.var))


@eager_propagate
def reduce_sum(
    data: _CoreArray, axes: _CoreArray, keepdims: int = 1, noop_with_empty_axes: int = 0
) -> _CoreArray:
    return _CoreArray(
        op.reduce_sum(
            data.var,
            axes.var,
            keepdims=keepdims,
            noop_with_empty_axes=noop_with_empty_axes,
        )
    )


@eager_propagate
def reduce_min(
    data: _CoreArray, axes: _CoreArray, keepdims: int = 1, noop_with_empty_axes: int = 0
) -> _CoreArray:
    return _CoreArray(
        op.reduce_min(
            data.var,
            axes.var,
            keepdims=keepdims,
            noop_with_empty_axes=noop_with_empty_axes,
        )
    )


@eager_propagate
def reduce_max(
    data: _CoreArray, axes: _CoreArray, keepdims: int = 1, noop_with_empty_axes: int = 0
) -> _CoreArray:
    return _CoreArray(
        op.reduce_max(
            data.var,
            axes.var,
            keepdims=keepdims,
            noop_with_empty_axes=noop_with_empty_axes,
        )
    )


@eager_propagate
def reduce_mean(
    data: _CoreArray, axes: _CoreArray, keepdims: int = 1, noop_with_empty_axes: int = 0
) -> _CoreArray:
    return _CoreArray(
        op.reduce_mean(
            data.var,
            axes.var,
            keepdims=keepdims,
            noop_with_empty_axes=noop_with_empty_axes,
        )
    )


@eager_propagate
def reduce_prod(
    data: _CoreArray, axes: _CoreArray, keepdims: int = 1, noop_with_empty_axes: int = 0
) -> _CoreArray:
    return _CoreArray(
        op.reduce_prod(
            data.var,
            axes.var,
            keepdims=keepdims,
            noop_with_empty_axes=noop_with_empty_axes,
        )
    )


@eager_propagate
def concat(inputs: list[_CoreArray], axis: int) -> _CoreArray:
    return _CoreArray(op.concat([x.var for x in inputs], axis=axis))


@eager_propagate
def unsqueeze(data: _CoreArray, axes: _CoreArray) -> _CoreArray:
    if axes.dtype != dtypes.int64:
        raise TypeError(f"axes must be int64, got {axes.dtype}")
    return _CoreArray(op.unsqueeze(data.var, axes.var))


@eager_propagate
def trilu(x: _CoreArray, *, k: _CoreArray | None = None, upper: int = 0) -> _CoreArray:
    return _CoreArray(op.trilu(x.var, k.var if k is not None else None, upper=upper))


@eager_propagate
def unique(
    x: _CoreArray, axis: int | None = None, sorted: int = 1
) -> tuple[_CoreArray, _CoreArray, _CoreArray, _CoreArray]:
    y, indices, inverse_indices, counts = op.unique(x.var, axis=axis, sorted=sorted)
    return (
        _CoreArray(y),
        _CoreArray(indices),
        _CoreArray(inverse_indices),
        _CoreArray(counts),
    )


@eager_propagate
def and_(x: _CoreArray, y: _CoreArray) -> _CoreArray:
    return _CoreArray(op.and_(x.var, y.var))


@eager_propagate
def gather(data: _CoreArray, indices: _CoreArray, axis: int = 0) -> _CoreArray:
    return _CoreArray(op.gather(data.var, indices.var, axis=axis))


@eager_propagate
def cumsum(
    x: _CoreArray, axis: _CoreArray, exclusive: int = 0, reverse: int = 0
) -> _CoreArray:
    return _CoreArray(
        op.cumsum(x.var, axis=axis.var, exclusive=exclusive, reverse=reverse)
    )


@eager_propagate
def split(
    input: _CoreArray,
    split: _CoreArray | None = None,
    *,
    num_outputs: int,
    axis: int = 0,
) -> tuple[_CoreArray, ...]:
    return tuple(
        map(
            _CoreArray,
            op.split(
                input.var,
                split.var if split is not None else None,
                num_outputs=num_outputs,
                axis=axis,
            ),
        )
    )


@eager_propagate
def clip(input: _CoreArray, min: _CoreArray, max: _CoreArray) -> _CoreArray:
    return _CoreArray(op.clip(input.var, min.var, max.var))


@eager_propagate
def matmul(a: _CoreArray, b: _CoreArray) -> _CoreArray:
    return _CoreArray(op.matmul(a.var, b.var))


@eager_propagate
def arg_max(
    data: _CoreArray, *, axis: int = 0, keepdims: int = 1, select_last_index: int = 0
) -> _CoreArray:
    return _CoreArray(
        op.arg_max(
            data.var, axis=axis, keepdims=keepdims, select_last_index=select_last_index
        )
    )


@eager_propagate
def arg_min(
    data: _CoreArray, *, axis: int = 0, keepdims: int = 1, select_last_index: int = 0
) -> _CoreArray:
    return _CoreArray(
        op.arg_min(
            data.var, axis=axis, keepdims=keepdims, select_last_index=select_last_index
        )
    )


@eager_propagate
def bitwise_and(x: _CoreArray, y: _CoreArray) -> _CoreArray:
    return _CoreArray(op.bitwise_and(x.var, y.var))


@eager_propagate
def bitwise_or(x: _CoreArray, y: _CoreArray) -> _CoreArray:
    return _CoreArray(op.bitwise_or(x.var, y.var))


@eager_propagate
def bitwise_xor(x: _CoreArray, y: _CoreArray) -> _CoreArray:
    return _CoreArray(op.bitwise_xor(x.var, y.var))


@eager_propagate
def bitwise_not(x: _CoreArray) -> _CoreArray:
    return _CoreArray(op.bitwise_not(x.var))


@eager_propagate
def bit_shift(x: _CoreArray, y: _CoreArray, *, direction: str) -> _CoreArray:
    return _CoreArray(op.bit_shift(x.var, y.var, direction=direction))


@eager_propagate
def xor(x: _CoreArray, y: _CoreArray) -> _CoreArray:
    return _CoreArray(op.xor(x.var, y.var))


@eager_propagate
def acosh(x: _CoreArray) -> _CoreArray:
    return _CoreArray(op.acosh(x.var))


@eager_propagate
def getitem_null(corearray: _CoreArray, index: _CoreArray) -> _CoreArray:
    # Compress doesn't compress in case of scalars
    # A solution that works is to transform
    # var -> [var]
    # index -> [index]

    var = corearray.var
    if get_rank(index) == 0:

        def extend_shape(x: Var) -> Var:
            return op.concat([op.const([1], dtype=np.int64), op.shape(x)], axis=0)

        var = op.reshape(var, extend_shape(var), allowzero=True)
        index_var = op.reshape(index.var, extend_shape(index.var), allowzero=True)
        return _CoreArray(op.compress(var, index_var, axis=0))

    else:
        if get_rank(index) == 1:
            reshaped_var, reshaped_index = var, index.var
        else:
            ret_shape = op.concat(
                [op.const([-1], dtype=np.int64), op.shape(var, start=get_rank(index))],
                axis=0,
            )
            reshaped_var = op.reshape(var, ret_shape, allowzero=True)
            reshaped_index = op.reshape(index.var, op.const([-1]), allowzero=True)

        return _CoreArray(
            op.compress(
                unsafe_reshape(
                    reshaped_var,
                    tuple(
                        None
                        for _ in builtins.range(get_rank(var) - get_rank(index) + 1)
                    ),
                ),
                reshaped_index,
                axis=0,
            )
        )


@eager_propagate
def getitem(
    corearray: _CoreArray,
    index: tuple[ScalarIndexType, ...] | _CoreArray,
) -> _CoreArray:
    if isinstance(index, _CoreArray):
        if get_dtype(index) == np.bool_:
            if get_rank(corearray) < get_rank(index):
                raise IndexError("Indexing with boolean array cannot happen")
            return getitem_null(corearray, index)
        else:
            return _CoreArray(op.gather(corearray.var, index.var, axis=0))
    elif len(index) == 0:
        return corearray.copy()
    elif all(isinstance(i, bool) for i in index):
        index = typing.cast(tuple[bool, ...], index)
        return getitem(
            corearray,
            _CoreArray(op.const(index) if len(index) > 1 else op.const(index[0])),
        )
    # index is a tuple of scalars
    index_some = [x for x in index if x is not None]

    axis_slices = [
        (x.start, x.stop, x.step, ind)
        for ind, x in enumerate(index_some)
        if isinstance(x, slice) and x.step is not None
    ]
    var = corearray.var
    if axis_slices:
        ndim = get_rank(var)
        var = op.slice(
            var,
            op.const(np.array([x[0] for x in axis_slices], np.int64)),
            op.const(np.array([x[1] for x in axis_slices], np.int64)),
            op.const(np.array([x[3] for x in axis_slices], np.int64)),
            op.const(np.array([x[2] for x in axis_slices], np.int64)),
        )

        # TODO: onnx should be fixing this
        var = unsafe_reshape(var, tuple(None for i in builtins.range(ndim)))

    axis_indices = [(ind, x) for ind, x in enumerate(index_some) if isinstance(x, int)]
    for axis, axis_index in reversed(axis_indices):
        var = op.gather(var, op.const(axis_index), axis=axis)

    index_filtered = [x for x in index if isinstance(x, (type(None), slice))]
    axis_new_axes = [ind for ind, x in enumerate(index_filtered) if x is None]
    if len(axis_new_axes) != 0:
        var = op.unsqueeze(var, axes=op.const(axis_new_axes, dtype=np.int64))

    return _CoreArray(var)


@eager_propagate
def setitem(
    corearray: _CoreArray, index: tuple | _CoreArray, updates: _CoreArray
) -> _CoreArray:
    if get_rank(corearray) == 0 and isinstance(index, tuple):
        return updates
    elif get_rank(corearray) == 0 and isinstance(index, _CoreArray):
        return _CoreArray(
            op.if_(
                index.var,
                then_branch=lambda: [updates.var],
                else_branch=lambda: [corearray.var],
            )[0]
        )

    var = corearray.var
    indices = getitem(ndindex(_CoreArray(op.shape(var))), index)

    # broadcast updates as appropriate
    index_path_shape = op.slice(
        op.shape(indices.var),
        op.const([0], dtype=np.int64),
        op.const([-1], dtype=np.int64),
    )
    return _CoreArray(
        op.scatter_nd(var, indices.var, op.expand(updates.var, index_path_shape))
    )


@eager_propagate
def ndindex(shape: _CoreArray, to_reverse=None, axes_permutation=None) -> _CoreArray:
    """
    Returns a tensor of shape `shape`, where each element is its index in the tensor
    Parameters:
        shape: the shape of the tensor
        to_reverse: a list of axes to reverse
        axes_permutation: a permutation of the axes
    Returns:
        A tensor of shape `shape`, where each element is its index in the tensor
    """
    if to_reverse is None:
        to_reverse = []
    (rank,) = get_shape(shape.var)

    if not isinstance(rank, int):
        raise ValueError("rank must be a constant integer")

    if rank == 0:
        return const([], dtype=dtypes.int64)

    if axes_permutation is None:
        axes_indices = list(builtins.range(rank))
    else:
        axes_indices = [axes_permutation.index(i) for i in builtins.range(rank)]

    shape_var = shape.var
    dtype = shape_var.unwrap_tensor().dtype
    ranges = [
        (
            op.range(
                op.const(0, dtype=dtype),
                op.gather(shape_var, op.const(i)),
                op.const(1, dtype=dtype),
            )
            if i not in to_reverse
            else op.range(
                op.sub(op.gather(shape_var, op.const(i)), op.const(1, dtype=dtype)),
                op.const(-1, dtype=dtype),
                op.const(-1, dtype=dtype),
            )
        )
        for i in builtins.range(rank)
    ]

    fit_ranges = [
        op.unsqueeze(
            r,
            op.const(
                [j for j in builtins.range(rank) if axes_indices[i] != j],
                dtype=np.int64,
            ),
        )
        for i, r in enumerate(ranges)
    ]

    if axes_permutation is not None:
        shape_var = op.gather(shape_var, op.const(axes_permutation))

    expanded_ranges = [op.expand(r, shape_var) for r in fit_ranges]

    ret = op.concat(
        [op.unsqueeze(r, op.const([-1], dtype=np.int64)) for r in expanded_ranges],
        axis=-1,
    )

    return _CoreArray(ret)


@eager_propagate
def reshape_like(x: _CoreArray, y: _CoreArray) -> _CoreArray:
    x_var, y_var = x.var, y.var
    return _CoreArray(
        unsafe_reshape(
            op.reshape(x_var, op.shape(y_var)),
            [x.to_simple() for x in y_var.unwrap_tensor()._shape],  # type: ignore
        )
    )


@eager_propagate
def static_map(
    input: _CoreArray, mapping: Mapping[KeyType, ValueType], default: ValueType | None
) -> _CoreArray:
    keys = np.asarray(tuple(mapping.keys()))
    values = np.asarray(tuple(mapping.values()))
    if isinstance(input.dtype, dtypes.Integral):
        input = input.astype(dtypes.int64)
        # Should only be a relevant path in Windows NumPy 1.x
        if keys.dtype != np.dtype("int64") and keys.dtype.kind == "i":
            keys = keys.astype(input.dtype.to_numpy_dtype())
    elif isinstance(input.dtype, dtypes.Floating):
        input = input.astype(dtypes.float64)

    value_dtype = values.dtype
    if default is None:
        if value_dtype.kind == "U" or (
            value_dtype.kind == "O" and all(isinstance(x, str) for x in values.flat)
        ):
            default_tensor = np.asarray(["MISSING"])
        else:
            default_tensor = np.asarray([0], dtype=value_dtype)
    elif value_dtype.kind == "U":
        default_tensor = np.asarray([default], dtype=np.str_)
    else:
        default_tensor = np.asarray([default], dtype=value_dtype)
    if keys.dtype.kind == "f" and isinstance(input.dtype, dtypes.Integral):
        input = cast(input, dtypes.from_numpy_dtype(keys.dtype))
    elif keys.dtype.kind == "i" and isinstance(input.dtype, dtypes.Floating):
        keys = keys.astype(input.dtype.to_numpy_dtype())
    return _CoreArray(
        ml.label_encoder(
            input.var,
            keys_tensor=keys,
            values_tensor=values,
            default_tensor=default_tensor,
        )
    )


@eager_propagate
def get_indices(
    x: _CoreArray, y: _CoreArray, positions: _CoreArray
) -> tuple[_CoreArray, _CoreArray]:
    """Given a position array representing the positions of x concatenated to those of
    y, this function extracts the original positions back out as two tensors."""
    split_pos = shape(x)
    split_pos_var = split_pos.var
    positions_var = positions.var
    indices_x = op.reshape(
        op.slice(positions_var, op.const([0], dtype=np.int64), split_pos_var),
        op.const([-1], dtype=np.int64),
    )
    indices_y = op.reshape(
        op.slice(positions_var, split_pos_var, op.const([np.iinfo(np.int64).max])),
        op.const([-1], dtype=np.int64),
    )
    return _CoreArray(indices_x), _CoreArray(indices_y)



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

import numpy as np

from ndonnx._funcs import asarray

e = asarray(np.e)
inf = asarray(np.inf)
nan = asarray(np.nan)
pi = asarray(np.pi)
newaxis = None

__all__ = [
    "e",
    "inf",
    "nan",
    "pi",
    "newaxis",
]



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

import numpy as np

import ndonnx as ndx
import ndonnx._core as impl
import ndonnx._opset_extensions as opx

from .conversion import CastError, CastMixin
from .schema import Schema

if TYPE_CHECKING:
    from ndonnx._array import Array

    from .structtype import StructType


class CoreType(CastMixin):
    """Base class for the standard data types present in ONNX."""

    @staticmethod
    @abstractmethod
    def to_numpy_dtype() -> np.dtype:
        """Returns the numpy dtype corresponding to this data type."""

    def __eq__(self, other: object) -> bool:
        return (
            type(self) is type(other)
            or type(self) is other
            or self.to_numpy_dtype() == other
        )

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __repr__(self) -> str:
        return type(self).__name__

    def __hash__(self) -> int:
        return hash(type(self))

    def __str__(self) -> str:
        return repr(self)

    def _schema(self) -> Schema:
        return Schema(type_name=type(self).__name__, author="ndonnx")

    def _assemble_output(self, fields: dict[str, np.ndarray]) -> np.ndarray:
        if "data" not in fields:
            raise ValueError("Fields must contain 'data' key")
        return fields["data"]

    def _parse_input(self, data: np.ndarray) -> dict[str, np.ndarray]:
        if data.dtype != self.to_numpy_dtype():
            raise ValueError(
                f"Expected dtype {self.to_numpy_dtype()}, got {data.dtype}"
            )
        if isinstance(data, np.ma.MaskedArray):
            raise ValueError(
                "Masked arrays are not supported for core data types. Use a nullable data type instead."
            )
        return {"data": data}

    def _cast_from(self, array: Array) -> Array:
        if isinstance(array.dtype, CoreType):
            return (
                ndx.Array._from_fields(self, data=opx.cast(array._core(), to=self))
                if array.dtype != self
                else array.copy()
            )
        else:
            raise CastError(f"Cannot cast from {array.dtype} to {self}")

    def _cast_to(self, array: Array, dtype: CoreType | StructType) -> Array:
        if isinstance(dtype, CoreType):
            return dtype._cast_from(array)
        else:
            raise CastError(f"Cannot cast from {array.dtype} to {dtype}")

    _ops: impl.OperationsBlock



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

import pytest

import ndonnx as ndx


def test_iter_for_loop():
    n = 5
    a = ndx.array(shape=(n,), dtype=ndx.int64)

    for i, el in enumerate(a):  # type: ignore
        assert isinstance(el, ndx.Array)
        if i > n:
            assert False, "Iterated past the number of elements"


@pytest.mark.parametrize(
    "arr",
    [
        ndx.asarray([1]),
        ndx.asarray([[1], [2]]),
        ndx.array(shape=(2,), dtype=ndx.int64),
        ndx.array(shape=(2, 3), dtype=ndx.int64),
        ndx.array(shape=(2, "N"), dtype=ndx.int64),
    ],
)
def test_create_iterators(arr):
    it = iter(arr)
    el = next(it)
    assert el.ndim == arr.ndim - 1
    assert el.shape == arr.shape[1:]


def test_0d_not_iterable():
    scalar = ndx.array(shape=(), dtype=ndx.int64)
    with pytest.raises(ValueError):
        next(iter(scalar))


def test_raises_dynamic_dim():
    with pytest.raises(ValueError):
        iter(ndx.array(shape=("N",), dtype=ndx.int64))



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

import sys

import numpy as np
import pytest

import ndonnx as ndx
import ndonnx.additional as nda

from .utils import assert_array_equal, run


@pytest.mark.parametrize(
    "side",
    [
        "left",
        "right",
    ],
)
def test_searchsorted(side):
    a_val = [0, 1, 2, 5, 5, 6, 10, 15]
    b_val = [0, 1, 2, 3, 4, 5, 6, 7, 8, 10, 10, 15, 20, 20]
    c_val = np.searchsorted(a_val, b_val, side)

    a = ndx.asarray(a_val, dtype=ndx.int64)
    b = ndx.asarray(b_val, dtype=ndx.int64)
    c = ndx.searchsorted(a, b, side=side)
    assert_array_equal(c_val, c.to_numpy())


@pytest.mark.skip(reason="TODO: onnxruntime")
@pytest.mark.parametrize(
    "side",
    [
        "left",
        "right",
    ],
)
def test_searchsorted_nans(side):
    a_val = np.array([0, 1, 2, 5, 5, 6, 10, 15, np.nan])
    b_val = np.array([0, 1, 2, np.nan, np.nan])
    c_val = np.searchsorted(a_val, b_val, side)

    a = ndx.array(shape=(len(a_val),), dtype=ndx.float64)
    b = ndx.array(shape=(len(b_val),), dtype=ndx.float64)
    c = ndx.searchsorted(a, b, side=side)

    model = ndx.build({"a": a, "b": b}, {"c": c})

    assert_array_equal(c_val, run(model, dict(a=a_val, b=b_val))["c"])


def test_searchsorted_raises():
    with pytest.raises(TypeError):
        a = ndx.array(shape=(), dtype=ndx.int64)
        b = ndx.array(shape=(), dtype=ndx.float64)

        ndx.searchsorted(a, b)

    with pytest.raises(ValueError):
        a = ndx.array(shape=(3,), dtype=ndx.int64)
        b = ndx.array(shape=(3,), dtype=ndx.int64)

        ndx.searchsorted(a, b, side="middle")  # type: ignore[arg-type]


@pytest.mark.skipif(
    sys.platform.startswith("win") and np.__version__ < "2",
    reason="ORT 1.18 not registering LabelEncoder(4) only on Windows.",
)
def test_static_map_lazy():
    a = ndx.array(shape=(3,), dtype=ndx.int64)
    b = nda.static_map(a, {1: 2, 2: 3})
    model = ndx.build({"a": a}, {"b": b})
    assert_array_equal([0, 2, 3], run(model, {"a": np.array([0, 1, 2])})["b"])

    # nans are mapped by static_map
    a = ndx.array(shape=("N",), dtype=ndx.float64)
    b = nda.static_map(a, {1.0: 2, np.nan: 3, 3.0: 42}, default=-1)

    model = ndx.build({"a": a}, {"b": b})
    assert_array_equal(
        [-1, -1, 42, 3],
        run(model, {"a": np.array([0.0, 2.0, 3.0, np.nan])})["b"],
    )


@pytest.mark.skipif(
    sys.platform.startswith("win") and np.__version__ < "2",
    reason="ORT 1.18 not registering LabelEncoder(4) only on Windows.",
)
@pytest.mark.parametrize(
    "x, mapping, default, expected",
    [
        (
            ndx.asarray(["hello", "world", "!"]),
            {"hello": "hi", "world": "earth"},
            None,
            ["hi", "earth", "MISSING"],
        ),
        (
            ndx.asarray(["hello", "world", "!"]),
            {"hello": "hi", "world": "earth"},
            "DIFFERENT",
            ["hi", "earth", "DIFFERENT"],
        ),
        (ndx.asarray([0, 1, 2], dtype=ndx.int64), {0: -1, 1: -2}, None, [-1, -2, 0]),
        (ndx.asarray([0, 1, 2], dtype=ndx.int64), {0: -1, 1: -2}, 42, [-1, -2, 42]),
        (
            ndx.asarray([[0], [1], [2]], dtype=ndx.int64),
            {0: -1, 1: -2},
            42,
            [[-1], [-2], [42]],
        ),
        (
            ndx.asarray([[0], [1], [2]], dtype=ndx.int32),
            {0: -1, 1: -2},
            42,
            [[-1], [-2], [42]],
        ),
        (
            ndx.asarray([[0], [1], [2]], dtype=ndx.int8),
            {0: -1, 1: -2},
            42,
            [[-1], [-2], [42]],
        ),
        (
            ndx.asarray([[0], [1], [2]], dtype=ndx.uint8),
            {0: -1, 1: -2},
            42,
            [[-1], [-2], [42]],
        ),
        (
            ndx.asarray([[0], [1], [np.nan]], dtype=ndx.float32),
            {0: -1, 1: -2, np.nan: 3.142},
            42,
            [[-1], [-2], [3.142]],
        ),
    ],
)
def test_static_map(x, mapping, default, expected):
    actual = nda.static_map(x, mapping, default=default)
    assert_array_equal(actual.to_numpy(), expected)


def test_static_map_unimplemented_for_nullable():
    a = ndx.asarray([1, 2, 3], dtype=ndx.int64)
    m = ndx.asarray([True, False, True])
    a = nda.make_nullable(a, m)

    with pytest.raises(ndx.UnsupportedOperationError):
        nda.static_map(a, {1: 2, 2: 3})


@pytest.mark.skipif(
    sys.platform.startswith("win"),
    reason="ORT 1.18 not registering LabelEncoder(4) only on Windows.",
)
def test_isin():
    a = ndx.array(shape=("N",), dtype=ndx.utf8)
    b = nda.isin(a, ["foo", "bar", "baz"])

    model = ndx.build({"a": a}, {"b": b})
    assert_array_equal(
        [False, True, True, False],
        run(model, dict(a=np.array(["hello", "foo", "baz", "!"])))["b"],
    )

    a = ndx.array(shape=("N",), dtype=ndx.utf8)
    b = nda.isin(a, ["🔴", "🟡", "🟢"])

    model = ndx.build({"a": a}, {"b": b})
    assert_array_equal(
        [False, True, False],
        run(model, dict(a=np.array(["🚀", "🔴", "hi🟡"])))["b"],
    )

    a = ndx.asarray(["hello", "world"])
    assert_array_equal([True, False], nda.isin(a, ["hello"]).to_numpy())


@pytest.mark.parametrize(
    "dtype",
    [
        ndx.int64,
        ndx.utf8,
        ndx.bool,
    ],
)
@pytest.mark.parametrize(
    "mask",
    [
        [True, False, True],
        True,
        False,
        [True],
    ],
)
def test_make_nullable(dtype, mask):
    a = ndx.asarray([1, 2, 3], dtype=dtype)
    m = ndx.asarray(mask)

    result = nda.make_nullable(a, m)
    expected = np.ma.masked_array([1, 2, 3], mask, dtype.to_numpy_dtype())
    assert_array_equal(result.to_numpy(), expected)



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

import numpy as np
import pytest

import ndonnx as ndx
import ndonnx._opset_extensions as opx

from .utils import assert_array_equal


@pytest.mark.parametrize("op", ["add", "multiply", "subtract", "divide"])
def test_eager_mode(op):
    a = ndx.asarray(np.array([0.0, 1, 2, 3]), dtype=ndx.float64)
    b = ndx.asarray(np.array([3.0, 4, 5, 6]), dtype=ndx.float64)
    assert_array_equal(
        getattr(ndx, op)(a, b).to_numpy(),
        getattr(np, op)(a.to_numpy(), b.to_numpy()),
    )


def test_indexing():
    a = ndx.asarray(np.array([0.0, 1, 2, 3]), dtype=ndx.float64)
    b = ndx.asarray(np.array([2, 1]), dtype=ndx.int64)
    assert_array_equal(a[b].to_numpy(), a.to_numpy()[b.to_numpy()])  # type: ignore
    a[1] = 3.4
    assert_array_equal(a.to_numpy(), np.array([0.0, 3.4, 2.0, 3.0], dtype=np.float64))
    a[2] = 4.2
    assert_array_equal(a.to_numpy(), np.array([0.0, 3.4, 4.2, 3.0], dtype=np.float64))
    a[b] = ndx.asarray(np.array([-1.0, np.nan]))
    assert_array_equal(
        a.to_numpy(), np.array([0.0, np.nan, -1.0, 3.0], dtype=np.float64)
    )


def test_slicing():
    a = ndx.asarray(np.array([0, 1, 2, 3], np.int64))
    b = ndx.asarray(np.array([1, 2], np.int64))
    c = a[1:3] * 2 + b
    assert_array_equal(c.to_numpy(), np.array([3, 6], dtype=np.int64))


def test_creation():
    a = ndx.asarray(np.array([0.0, 1, 2, 3]))
    assert a.to_numpy() is not None
    assert_array_equal(a.to_numpy(), np.array([0.0, 1, 2, 3], dtype=np.float64))


def test_opset_extensions_eager_propagation():
    a = ndx.asarray(
        np.arange(0, 4, dtype=np.int64),
    )

    b = ndx.asarray(
        np.arange(10, 14, dtype=np.int64),
    )
    result = opx.add(a.data, b.data)  # type: ignore
    assert_array_equal(result.to_numpy(), np.asarray([10, 12, 14, 16], dtype=np.int64))
    result = opx.split(opx.concat([a.data, b.data], axis=0), num_outputs=2, axis=0)  # type: ignore
    assert isinstance(result, tuple) and len(result) == 2
    assert_array_equal(result[0].to_numpy(), a.to_numpy())
    assert_array_equal(result[1].to_numpy(), b.to_numpy())


@pytest.mark.skip(
    reason="Opset dispatch constant propagation leads to slow build times."
)
def test_opset_extensions_constant_folding():
    a = ndx.asarray(np.array([1, 2, 3]))
    b = ndx.asarray(np.array([4, 5, 6]))
    result = opx.add(a, b)
    assert result._var._op.op_type.identifier == "Constant"
    assert_array_equal(result.to_numpy(), [5, 7, 9])


def test_eager_propagate_decorator():
    import ndonnx._opset_extensions as opx
    from ndonnx._array import _from_corearray
    from ndonnx._propagation import eager_propagate

    @eager_propagate
    def sin_plus(a, b):
        return _from_corearray(opx.sin(a.data)) + b

    a = ndx.asarray(
        np.array([0, 1, 2, 3], dtype=np.float64),
    )
    b = ndx.asarray(
        np.array([10, 11, 12, 13], dtype=np.float64),
    )
    result = sin_plus(a, b)
    assert_array_equal(result.to_numpy(), np.sin(a.to_numpy()) + b.to_numpy())

    @eager_propagate
    def assign_inplace(a, indices, values):
        a[indices] = values
        return a

    index = ndx.asarray(np.array([0, 3], dtype=np.int64))
    c = ndx.asarray(np.array([100, 200], dtype=np.int64))
    result = assign_inplace(a, index, c)
    assert_array_equal(result.to_numpy(), [100.0, 1, 2, 200])
    assert_array_equal(a.to_numpy(), [100.0, 1, 2, 200])
    assert a is result



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import dataclasses
import functools
import json

import numpy as np
import onnx
import spox

import ndonnx as ndx
import ndonnx._data_types as dtypes

from ._array import Array
from ._corearray import _CoreArray


def build(
    inputs: dict[str, Array],
    outputs: dict[str, Array],
) -> onnx.ModelProto:
    """Build utility to produce an ONNX model. Requires a dictionary of named input and
    output arrays.

    Parameters
    ----------
    inputs: dict[str, Array]
        Inputs of the model
    outputs: dict[str, Array]
        Outputs of the model

    Returns
    -------
    out: onnx.ModelProto
        ONNX model
    """

    def collect_vars(name: str, arr: _CoreArray | Array):
        if isinstance(arr, _CoreArray):
            return {name: arr.var}
        elif isinstance(arr.dtype, ndx.CoreType):
            return {name: arr.spox_var()}
        else:
            constituent_vars: dict[str, spox.Var] = {}
            for field_name, field in arr._fields.items():
                constituent_vars |= collect_vars(f"{name}_{field_name}", field)
            return constituent_vars

    input_vars: dict[str, spox.Var] = functools.reduce(
        lambda acc, arr: acc | collect_vars(*arr), inputs.items(), {}
    )
    output_vars: dict[str, spox.Var] = functools.reduce(
        lambda acc, arr: acc | collect_vars(*arr), outputs.items(), {}
    )
    model_proto = spox.build(input_vars, output_vars)
    input_schema = {
        name: dataclasses.asdict(input.dtype._schema())
        for name, input in inputs.items()
    }
    output_schema = {
        name: dataclasses.asdict(output.dtype._schema())
        for name, output in outputs.items()
    }
    model_proto.metadata_props.append(
        onnx.StringStringEntryProto(
            key="ndonnx_schema",
            value=json.dumps(
                {
                    "version": 1,
                    "input_schema": input_schema,
                    "output_schema": output_schema,
                }
            ),
        )
    )
    return model_proto


def _get_dtype(dtype: str, version: int) -> dtypes.StructType | dtypes.CoreType:
    """Get the ndonnx data type given a string representation and version number, as
    found in the produced `onnx.ModelProto`."""
    if version == 1:
        return _v1_dtypes[dtype]
    else:
        raise ValueError(f"Unknown schema version {version}")


def _extract_output_names(
    output_dtypes: dict[str, dtypes.StructType | dtypes.CoreType],
) -> dict[str, ndx.CoreType]:
    """Given a dictionary mapping output names to their data types, extract the
    underlying fully qualified names and their CoreTypes.

    Parameters
    ----------
    output_dtypes: dict[str, ndonnx.StructType | ndonnx.CoreType]
        Dictionary mapping output names to their data types

    Returns
    -------
    out: dict[str, ndonnx.CoreType]
        Dictionary mapping fully qualified output names to their CoreTypes
    """

    def extract_output_name(name, type):
        if isinstance(type, ndx.CoreType):
            return {name: type}
        else:
            names: dict[str, ndx.CoreType] = {}
            for field_name, field_type in type._fields().items():
                names |= extract_output_name(f"{name}_{field_name}", field_type)
            return names

    return functools.reduce(
        lambda acc, arr: acc | extract_output_name(*arr), output_dtypes.items(), {}
    )


def _flatten(input_dict, dtype: dtypes.CoreType | dtypes.StructType, field_name: str):
    if isinstance(dtype, ndx.CoreType):
        return {field_name: input_dict["data"]}
    else:
        flattened: dict[str, np.ndarray] = {}
        for field, field_type in dtype._fields().items():
            flattened |= _flatten(
                input_dict[field], field_type, f"{field_name}_{field}"
            )
        return flattened


def _deconstruct_inputs(
    inputs: dict[str, np.ndarray],
    input_schema: dict[str, dtypes.CoreType | dtypes.StructType],
) -> dict[str, np.ndarray]:
    """Given a dictionary of named inputs and a schema mapping these input names to
    their ndonnx data types, return the disassembled inputs as valid inputs for
    inference.

    Parameters
    ----------
    inputs: dict[str, np.ndarray]
        Dictionary of named inputs
    input_schema: dict[str, ndonnx.CoreType | ndonnx.StructType]
        Schema mapping input names to their ndonnx data types

    Returns
    -------
    out: dict[str, np.ndarray]
        Disassembled inputs with fully qualified names of core constituent fields of the inputs, based on the data type of the input in `input_schema`, based on the data type's `parse_input` method.
    """

    def parse_input(name, type):
        structured_fields = type._parse_input(inputs[name])
        flattened_fields = _flatten(structured_fields, input_schema[name], name)
        return flattened_fields

    return functools.reduce(
        lambda acc, arr: acc | parse_input(*arr), input_schema.items(), {}
    )


def _assemble_outputs(
    output_data: dict[str, np.ndarray],
    output_schema: dict[str, dtypes.CoreType | dtypes.StructType],
) -> dict[str, np.ndarray]:
    """Given a dictionary of named outputs and a schema mapping these output names to
    their ndonnx data types, return the outputs assembled into values that reflect their
    data type.

    Parameters
    ----------
    output_data: dict[str, np.ndarray]
        Dictionary of named outputs from inference
    output_schema: dict[str, ndonnx.CoreType | ndonnx.StructType]
        Schema mapping output names to their ndonnx data types

    Returns
    -------
    out: dict[str, np.ndarray]
        Assembled outputs with data type that is meaningful based on the `output_schema` entry, based on the data type's `assemble_output` method.
    """

    def _assemble_output(name, type):
        # We want to recurse down the type here and recursively reassemble the output
        def helper(cur_type, prefix):
            if isinstance(cur_type, ndx.CoreType):
                return cur_type._assemble_output({"data": output_data[prefix]})
            elif isinstance(cur_type, dtypes.StructType):
                inputs = {
                    field: helper(field_type, f"{prefix}_{field}")
                    for field, field_type in cur_type._fields().items()
                }
                return cur_type._assemble_output(inputs)
            else:
                raise TypeError

        return {name: helper(type, name)}

    return functools.reduce(
        lambda acc, arr: acc | _assemble_output(*arr), output_schema.items(), {}
    )


_v1_dtypes: dict[str, dtypes.CoreType | dtypes.StructType] = {
    "UInt8": dtypes.uint8,
    "UInt16": dtypes.uint16,
    "UInt32": dtypes.uint32,
    "UInt64": dtypes.uint64,
    "Int8": dtypes.int8,
    "Int16": dtypes.int16,
    "Int32": dtypes.int32,
    "Int64": dtypes.int64,
    "Float32": dtypes.float32,
    "Float64": dtypes.float64,
    "Utf8": dtypes.utf8,
    "Boolean": dtypes.bool,
    "NUInt8": dtypes.nuint8,
    "NUInt16": dtypes.nuint16,
    "NUInt32": dtypes.nuint32,
    "NUInt64": dtypes.nuint64,
    "NInt8": dtypes.nint8,
    "NInt16": dtypes.nint16,
    "NInt32": dtypes.nint32,
    "NInt64": dtypes.nint64,
    "NFloat32": dtypes.nfloat32,
    "NFloat64": dtypes.nfloat64,
    "NUtf8": dtypes.nutf8,
    "NBoolean": dtypes.nbool,
}



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause
from __future__ import annotations

import functools
import itertools
from typing import TYPE_CHECKING

import numpy as np

import ndonnx as ndx
import ndonnx._data_types as dtypes
from ndonnx._utility import promote

if TYPE_CHECKING:
    from ndonnx import Array
    from ndonnx._data_types import Dtype


def binary_op(
    x,
    y,
    op,
    via_dtype=None,
):
    return variadic_op([x, y], op, via_dtype)


def unary_op(
    x,
    op,
    via_dtype=None,
):
    return variadic_op([x], op, via_dtype)


def variadic_op(
    args,
    op,
    via_dtype=None,
    cast_return=True,
):
    args = promote(*args)
    out_dtype = args[0].dtype
    if not isinstance(out_dtype, (dtypes.CoreType, dtypes.NullableCore)):
        raise TypeError(
            f"Expected ndx.Array with CoreType or NullableCoreType, got {args[0].dtype}"
        )
    data, nulls = split_nulls_and_values(*args)
    if via_dtype is None:
        values = from_corearray(op(*(x._core() for x in data)))
    else:
        values = _via_dtype(op, via_dtype, data, cast_return=cast_return)

    if (out_null := functools.reduce(_or_nulls, nulls)) is not None:
        dtype = dtypes.into_nullable(values.dtype)
        return ndx.Array._from_fields(dtype, values=values, null=out_null)
    else:
        return values


def split_nulls_and_values(*xs):
    """Helper function that splits a series of ndx.Arrays into their constituent value
    and null mask components.

    Raises if an unexpected typed array is provided.
    """
    data: list[Array] = []
    nulls: list[Array | None] = []
    for x in xs:
        if isinstance(x.dtype, dtypes.Nullable):
            nulls.append(x.null)
            data.append(x.values)
        elif isinstance(x, ndx.Array):
            nulls.append(None)
            data.append(x)
        else:
            raise TypeError(f"Expected ndx.Array got {x}")
    return data, nulls


def _or_nulls(x, y):
    if x is None:
        return y
    if y is None:
        return x
    return ndx.logical_or(x, y)


def _via_dtype(
    fn,
    dtype,
    arrays,
    *,
    cast_return=True,
):
    """Call ``fn`` after casting to ``dtype`` and cast result back to the input dtype.

    The point of this function is to work around quirks / limited type support in the
    onnxruntime.

    ndx.Arrays are promoted to a common type prior to their first use.
    """
    promoted = promote(*arrays)
    out_dtype = promoted[0].dtype

    if isinstance(out_dtype, dtypes.NullableCore) and out_dtype.values == dtype:
        dtype = out_dtype

    values, nulls = split_nulls_and_values(
        *[ndx.astype(arr, dtype) for arr in promoted]
    )

    out_value = from_corearray(fn(*[value._core() for value in values]))
    out_null = functools.reduce(_or_nulls, nulls)

    if out_null is not None:
        out_value = ndx.Array._from_fields(
            dtypes.into_nullable(out_value.dtype), values=out_value, null=out_null
        )
    else:
        out_value = ndx.Array._from_fields(out_value.dtype, data=out_value._core())

    if cast_return:
        return ndx.astype(out_value, dtype=out_dtype)
    else:
        return out_value


def via_upcast(
    fn,
    arrays,
    *,
    int_dtype=None,
    float_dtype=None,
    uint_dtype=None,
    use_unsafe_uint_cast=False,
    cast_return=True,
):
    """Like ``_via_dtype`` but chooses a dtype from the available that doesn't result in
    loss.

    Raises
    ------
    TypeError
        For non-numerical types and if the type can't be safely casted to
        any available type.
    """
    promoted_values = promote(*arrays)

    dtype = promoted_values[0].dtype

    # For logging
    available_types = [t for t in [int_dtype, float_dtype, uint_dtype] if t is not None]

    via_dtype: dtypes.CoreType
    if isinstance(dtype, (dtypes.Unsigned, dtypes.NullableUnsigned)):
        if (
            uint_dtype is not None
            and ndx.iinfo(dtype).bits <= ndx.iinfo(uint_dtype).bits
        ):
            via_dtype = uint_dtype
        elif int_dtype is not None and ndx.iinfo(dtype).bits <= (
            ndx.iinfo(int_dtype).bits
            if use_unsafe_uint_cast
            else ndx.iinfo(int_dtype).bits - 1
        ):
            via_dtype = int_dtype
        else:
            raise TypeError(
                f"Can't upcast unsigned type `{dtype}`. Available implementations are for `{*available_types,}`"
            )
    elif isinstance(dtype, (dtypes.Integral, dtypes.NullableIntegral)) or dtype in (
        dtypes.nbool,
        dtypes.bool,
    ):
        if int_dtype is not None and ndx.iinfo(dtype).bits <= ndx.iinfo(int_dtype).bits:
            via_dtype = int_dtype
        else:
            raise TypeError(
                f"Can't upcast signed type `{dtype}`. Available implementations are for `{*available_types,}`"
            )
    elif isinstance(dtype, (dtypes.Floating, dtypes.NullableFloating)):
        if (
            float_dtype is not None
            and ndx.finfo(dtype).bits <= ndx.finfo(float_dtype).bits
        ):
            via_dtype = float_dtype
        else:
            raise TypeError(
                f"Can't upcast float type `{dtype}`. Available implementations are for `{*available_types,}`"
            )
    else:
        raise TypeError(f"Expected numerical data type, found {dtype}")

    return variadic_op(arrays, fn, via_dtype, cast_return)


def from_corearray(
    corearray,
):
    return ndx.Array._from_fields(corearray.dtype, data=corearray)


def validate_core(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        for arg in itertools.chain(args, kwargs.values()):
            if isinstance(arg, ndx.Array) and not isinstance(
                arg.dtype, (dtypes.CoreType, dtypes.NullableCore)
            ):
                return NotImplemented
        return func(*args, **kwargs)

    return wrapper


def assemble_output_recurse(dtype: Dtype, values: dict) -> np.ndarray:
    if isinstance(dtype, dtypes.CoreType):
        return dtype._assemble_output(values)
    else:
        fields = {
            name: assemble_output_recurse(field_dtype, values[name])
            for name, field_dtype in dtype._fields().items()
        }
        return dtype._assemble_output(fields)



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

import importlib.metadata
import warnings

from ._array import Array, array, from_spox_var
from ._build import (
    build,
)
from ._data_types import (
    CastError,
    CoreType,
    Floating,
    Integral,
    Nullable,
    NullableCore,
    NullableFloating,
    NullableIntegral,
    NullableNumerical,
    Numerical,
    promote_nullable,
    from_numpy_dtype,
    bool,
    float32,
    float64,
    int8,
    int16,
    int32,
    int64,
    nbool,
    nfloat32,
    nfloat64,
    nint8,
    nint16,
    nint32,
    nint64,
    nuint8,
    nuint16,
    nuint32,
    nuint64,
    nutf8,
    uint8,
    uint16,
    uint32,
    uint64,
    utf8,
)
from ._funcs import (
    arange,
    asarray,
    empty,
    empty_like,
    eye,
    full,
    full_like,
    linspace,
    ones,
    ones_like,
    tril,
    triu,
    zeros,
    zeros_like,
    astype,
    broadcast_arrays,
    broadcast_to,
    can_cast,
    finfo,
    iinfo,
    result_type,
    abs,
    acos,
    acosh,
    add,
    asin,
    asinh,
    atan,
    atan2,
    atanh,
    bitwise_and,
    bitwise_left_shift,
    bitwise_invert,
    bitwise_or,
    bitwise_right_shift,
    bitwise_xor,
    ceil,
    cos,
    cosh,
    divide,
    equal,
    exp,
    expm1,
    floor,
    floor_divide,
    greater,
    greater_equal,
    isdtype,
    isfinite,
    isinf,
    isnan,
    less,
    less_equal,
    log,
    log1p,
    log2,
    log10,
    logaddexp,
    logical_and,
    logical_not,
    logical_or,
    logical_xor,
    multiply,
    negative,
    not_equal,
    positive,
    pow,
    remainder,
    round,
    sign,
    sin,
    sinh,
    square,
    sqrt,
    subtract,
    tan,
    tanh,
    trunc,
    matmul,
    matrix_transpose,
    concat,
    expand_dims,
    flip,
    permute_dims,
    reshape,
    roll,
    squeeze,
    stack,
    argmax,
    argmin,
    nonzero,
    searchsorted,
    where,
    unique_all,
    unique_counts,
    unique_inverse,
    unique_values,
    argsort,
    sort,
    cumulative_sum,
    max,
    mean,
    min,
    prod,
    clip,
    std,
    sum,
    var,
    all,
    any,
    take,
    UnsupportedOperationError,
)
from ._constants import (
    e,
    inf,
    nan,
    pi,
)

try:
    __version__ = importlib.metadata.version(__name__)
except importlib.metadata.PackageNotFoundError as err:  # pragma: no cover
    warnings.warn(f"Could not determine version of {__name__}\n{err!s}", stacklevel=2)
    __version__ = "unknown"


__all__ = [
    "Array",
    "array",
    "from_spox_var",
    "e",
    "inf",
    "nan",
    "pi",
    "arange",
    "asarray",
    "empty",
    "empty_like",
    "eye",
    "full",
    "full_like",
    "linspace",
    "ones",
    "ones_like",
    "tril",
    "triu",
    "zeros",
    "zeros_like",
    "astype",
    "take",
    "broadcast_arrays",
    "broadcast_to",
    "can_cast",
    "finfo",
    "iinfo",
    "result_type",
    "abs",
    "acos",
    "acosh",
    "add",
    "asin",
    "asinh",
    "atan",
    "atan2",
    "atanh",
    "bitwise_and",
    "bitwise_left_shift",
    "bitwise_invert",
    "bitwise_or",
    "bitwise_right_shift",
    "bitwise_xor",
    "ceil",
    "cos",
    "cosh",
    "divide",
    "equal",
    "exp",
    "expm1",
    "floor",
    "floor_divide",
    "greater",
    "greater_equal",
    "isdtype",
    "isfinite",
    "isinf",
    "isnan",
    "less",
    "less_equal",
    "log",
    "log1p",
    "log2",
    "log10",
    "logaddexp",
    "logical_and",
    "logical_not",
    "logical_or",
    "logical_xor",
    "multiply",
    "negative",
    "not_equal",
    "positive",
    "pow",
    "remainder",
    "round",
    "sign",
    "sin",
    "sinh",
    "square",
    "sqrt",
    "subtract",
    "tan",
    "tanh",
    "trunc",
    "matmul",
    "matrix_transpose",
    "concat",
    "expand_dims",
    "flip",
    "permute_dims",
    "reshape",
    "roll",
    "squeeze",
    "stack",
    "argmax",
    "argmin",
    "nonzero",
    "searchsorted",
    "where",
    "unique_all",
    "unique_counts",
    "unique_inverse",
    "unique_values",
    "argsort",
    "sort",
    "cumulative_sum",
    "max",
    "mean",
    "min",
    "prod",
    "clip",
    "std",
    "sum",
    "var",
    "all",
    "any",
    "build",
    "bool",
    "utf8",
    "float32",
    "float64",
    "int8",
    "int16",
    "int32",
    "int64",
    "uint8",
    "uint16",
    "uint32",
    "uint64",
    "nbool",
    "nutf8",
    "nfloat32",
    "nfloat64",
    "nint8",
    "nint16",
    "nint32",
    "nint64",
    "nuint8",
    "nuint16",
    "nuint32",
    "nuint64",
    "NullableNumerical",
    "Numerical",
    "NullableFloating",
    "Floating",
    "NullableIntegral",
    "Nullable",
    "NullableCore",
    "Integral",
    "CoreType",
    "CastError",
    "promote_nullable",
    "from_numpy_dtype",
    "UnsupportedOperationError",
]



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

import ndonnx as ndx
import ndonnx._data_types as dtypes
import ndonnx._opset_extensions as opx

from ._coreimpl import CoreOperationsImpl
from ._interface import OperationsBlock
from ._nullableimpl import NullableOperationsImpl
from ._utils import binary_op, validate_core

if TYPE_CHECKING:
    from ndonnx import Array


class _StringOperationsImpl(OperationsBlock):
    @validate_core
    def add(self, x, y) -> Array:
        return binary_op(x, y, opx.string_concat)

    @validate_core
    def equal(self, x, y) -> Array:
        return binary_op(x, y, opx.equal)

    @validate_core
    def not_equal(self, x, y) -> ndx.Array:
        return ndx.logical_not(self.equal(x, y))

    @validate_core
    def can_cast(self, from_, to) -> bool:
        if isinstance(from_, ndx.CoreType) and isinstance(to, ndx.CoreType):
            return np.can_cast(from_.to_numpy_dtype(), to.to_numpy_dtype())
        return NotImplemented

    @validate_core
    def zeros(
        self,
        shape,
        dtype: dtypes.CoreType | dtypes.StructType | None = None,
        device=None,
    ):
        return ndx.full(shape, "", dtype=dtype)

    @validate_core
    def zeros_like(
        self, x, dtype: dtypes.CoreType | dtypes.StructType | None = None, device=None
    ):
        if dtype is not None and not isinstance(
            dtype, (dtypes.CoreType, dtypes.NullableCore)
        ):
            raise TypeError("'dtype' must be a CoreType or NullableCoreType")
        if dtype in (None, dtypes.utf8, dtypes.nutf8):
            return ndx.full_like(x, "", dtype=dtype)
        else:
            return ndx.full_like(x, 0, dtype=dtype)

    @validate_core
    def empty(self, shape, dtype=None, device=None) -> ndx.Array:
        return ndx.zeros(shape, dtype=dtype, device=device)

    @validate_core
    def empty_like(self, x, dtype=None, device=None) -> ndx.Array:
        return ndx.zeros_like(x, dtype=dtype, device=device)


class StringOperationsImpl(CoreOperationsImpl, _StringOperationsImpl): ...


class NullableStringOperationsImpl(NullableOperationsImpl, _StringOperationsImpl): ...



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Generic, TypeVar

import numpy as np
from typing_extensions import Self

import ndonnx as ndx
from ndonnx._core import (
    BooleanOperationsImpl,
    NullableBooleanOperationsImpl,
    NullableNumericOperationsImpl,
    NullableStringOperationsImpl,
    NumericOperationsImpl,
    OperationsBlock,
    StringOperationsImpl,
)
from ndonnx._data_types.schema import Schema

from .conversion import CastError, CastMixin
from .coretype import CoreType
from .structtype import StructType

if TYPE_CHECKING:
    from ndonnx import Array


class Numerical(CoreType):
    """Base class for numerical data types."""

    _ops: OperationsBlock = NumericOperationsImpl()


class Integral(Numerical):
    """Base class for integral data types."""


class Unsigned(Integral):
    """Base class for unsigned integral data types."""


class Fractional(Numerical):
    """Base class for fractional data types."""


class Floating(Fractional):
    """Base class for floating data types."""


class Int8(Integral):
    """8-bit signed integer."""

    @staticmethod
    def to_numpy_dtype() -> np.dtype:
        return np.dtype("int8")


class Int16(Integral):
    """16-bit signed integer."""

    @staticmethod
    def to_numpy_dtype() -> np.dtype:
        return np.dtype("int16")


class Int32(Integral):
    """32-bit signed integer."""

    @staticmethod
    def to_numpy_dtype() -> np.dtype:
        return np.dtype("int32")


class Int64(Integral):
    """64-bit signed integer."""

    @staticmethod
    def to_numpy_dtype() -> np.dtype:
        return np.dtype("int64")


class UInt8(Unsigned):
    """8-bit unsigned integer."""

    @staticmethod
    def to_numpy_dtype() -> np.dtype:
        return np.dtype("uint8")


class UInt16(Unsigned):
    """16-bit unsigned integer."""

    @staticmethod
    def to_numpy_dtype() -> np.dtype:
        return np.dtype("uint16")


class UInt32(Unsigned):
    """32-bit unsigned integer."""

    @staticmethod
    def to_numpy_dtype() -> np.dtype:
        return np.dtype("uint32")


class UInt64(Unsigned):
    """64-bit unsigned integer."""

    @staticmethod
    def to_numpy_dtype() -> np.dtype:
        return np.dtype("uint64")


class Float32(Floating):
    """32-bit floating point."""

    @staticmethod
    def to_numpy_dtype() -> np.dtype:
        return np.dtype("float32")


class Float64(Floating):
    """64-bit floating point."""

    @staticmethod
    def to_numpy_dtype() -> np.dtype:
        return np.dtype("float64")


class Boolean(CoreType):
    """Boolean type."""

    @staticmethod
    def to_numpy_dtype() -> np.dtype:
        return np.dtype("bool")

    _ops: OperationsBlock = BooleanOperationsImpl()


class Utf8(CoreType):
    """UTF-8 encoded string."""

    @staticmethod
    def to_numpy_dtype() -> np.dtype:
        return np.dtype("str")

    def _parse_input(self, data: np.ndarray) -> dict[str, np.ndarray]:
        if data.dtype.kind == "U" or (
            data.dtype.kind == "O" and all(isinstance(x, str) for x in data.flat)
        ):
            return {"data": data.astype(np.str_)}
        else:
            raise TypeError(f"Expected data type with kind 'U', got {data.dtype}")

    def _assemble_output(self, fields: dict[str, np.ndarray]) -> np.ndarray:
        if "data" not in fields:
            raise ValueError(
                f"Missing fields in output: {fields.keys()}, expected `data`"
            )
        if fields["data"].dtype.kind == "O" and all(
            isinstance(x, str) for x in fields["data"]
        ):
            return fields["data"].astype(np.str_)
        elif fields["data"].dtype.kind == "U":
            return fields["data"]
        else:
            raise TypeError(
                f"Expected data type with kind 'U', got {fields['data'].dtype}"
            )

    _ops: OperationsBlock = StringOperationsImpl()


T = TypeVar("T", StructType, CoreType, covariant=True)


class Nullable(StructType, Generic[T]):
    values: T
    null: Boolean

    def _fields(self) -> dict[str, StructType | CoreType]:
        return {
            "values": self.values,
            "null": self.null,
        }


class NullableCore(Nullable[CoreType], CastMixin):
    def copy(self) -> Self:
        return self

    def _parse_input(self, x: np.ndarray) -> dict:
        if not isinstance(x, np.ma.MaskedArray):
            raise TypeError(f"Expected numpy MaskedArray, got {type(x)}")
        return {
            "values": self.values._parse_input(x.data),
            "null": self.null._parse_input(x.mask),
        }

    def _assemble_output(self, fields: dict[str, np.ndarray]) -> np.ndarray:
        if "values" not in fields or "null" not in fields:
            raise ValueError(
                f"Missing fields in output: {fields.keys()}, expected `null` and `values`"
            )

        return np.ma.masked_array(fields["values"], mask=fields["null"])

    def _schema(self) -> Schema:
        return Schema(type_name=type(self).__name__, author="ndonnx")

    def _cast_to(self, array: Array, dtype: CoreType | StructType) -> Array:
        if isinstance(dtype, NullableCore):
            return ndx.Array._from_fields(
                dtype,
                values=self.values._cast_to(array.values, dtype.values),
                null=array.null.copy(),
            )
        else:
            raise CastError(f"Cannot cast to {dtype} from {self}")

    def _cast_from(self, array: Array) -> Array:
        if isinstance(array.dtype, CoreType):
            # Promote to nullable variant
            return ndx.Array._from_fields(
                self,
                values=self.values._cast_from(array),
                null=ndx.zeros_like(array, dtype=Boolean()),
            )
        elif isinstance(array.dtype, NullableCore):
            return ndx.Array._from_fields(
                self,
                values=self.values._cast_from(array.values),
                null=array.null.copy(),
            )
        else:
            raise CastError(f"Cannot cast from {array.dtype} to {self}")


class NullableNumerical(NullableCore):
    """Base class for nullable numerical data types."""

    _ops: OperationsBlock = NullableNumericOperationsImpl()


class NullableIntegral(NullableNumerical):
    """Base class for nullable integral data types."""


class NullableUnsigned(NullableIntegral):
    """Base class for nullable unsigned integral data types."""


class NullableFractional(NullableNumerical):
    """Base class for nullable fractional data types."""


class NullableFloating(NullableFractional):
    """Base class for nullable floating data types."""


class NInt8(NullableIntegral):
    values = Int8()
    null = Boolean()


class NInt16(NullableIntegral):
    values = Int16()
    null = Boolean()


class NInt32(NullableIntegral):
    values = Int32()
    null = Boolean()


class NInt64(NullableIntegral):
    values = Int64()
    null = Boolean()


class NUInt8(NullableUnsigned):
    values = UInt8()
    null = Boolean()


class NUInt16(NullableUnsigned):
    values = UInt16()
    null = Boolean()


class NUInt32(NullableUnsigned):
    values = UInt32()
    null = Boolean()


class NUInt64(NullableUnsigned):
    values = UInt64()
    null = Boolean()


class NFloat32(NullableFloating):
    values = Float32()
    null = Boolean()


class NFloat64(NullableFloating):
    values = Float64()
    null = Boolean()


class NBoolean(NullableCore):
    values = Boolean()
    null = Boolean()

    _ops: OperationsBlock = NullableBooleanOperationsImpl()


class NUtf8(NullableCore):
    values = Utf8()
    null = Boolean()

    _ops: OperationsBlock = NullableStringOperationsImpl()


def from_numpy_dtype(np_dtype: np.dtype) -> CoreType:
    """Convert a numpy dtype to an ndonnx data type."""
    if np_dtype == np.dtype("int8"):
        return Int8()
    if np_dtype == np.dtype("int16"):
        return Int16()
    if np_dtype == np.dtype("int32"):
        return Int32()
    if np_dtype == np.dtype("int64"):
        return Int64()
    if np_dtype == np.dtype("uint8"):
        return UInt8()
    if np_dtype == np.dtype("uint16"):
        return UInt16()
    if np_dtype == np.dtype("uint32"):
        return UInt32()
    if np_dtype == np.dtype("uint64"):
        return UInt64()
    if np_dtype == np.dtype("float32"):
        return Float32()
    if np_dtype == np.dtype("float64"):
        return Float64()
    if np_dtype == np.dtype("bool"):
        return Boolean()
    if np_dtype == np.dtype("str") or np_dtype.kind in ("U", "O"):
        return Utf8()
    raise TypeError(f"Unsupported data type: {np_dtype}")


@dataclass
class Iinfo:
    # Number of bits occupied by the type.
    bits: int
    # Largest representable number.
    max: int
    # Smallest representable number.
    min: int
    # Integer data type.
    dtype: CoreType

    @classmethod
    def _from_dtype(cls, dtype: CoreType) -> Iinfo:
        iinfo = np.iinfo(dtype.to_numpy_dtype())
        return cls(
            bits=iinfo.bits,
            max=iinfo.max,
            min=iinfo.min,
            dtype=dtype,
        )


@dataclass
class Finfo:
    # number of bits occupied by the real-valued floating-point data type.
    bits: int
    # difference between 1.0 and the next smallest representable real-valued floating-point number larger than 1.0 according to the IEEE-754 standard.
    eps: float
    # largest representable real-valued number.
    max: float
    # smallest representable real-valued number.
    min: float
    # smallest positive real-valued floating-point number with full precision.
    smallest_normal: float
    # real-valued floating-point data type.
    dtype: CoreType

    @classmethod
    def _from_dtype(cls, dtype: CoreType) -> Finfo:
        finfo = np.finfo(dtype.to_numpy_dtype())
        return cls(
            bits=finfo.bits,
            max=float(finfo.max),
            min=float(finfo.min),
            dtype=dtype,
            eps=float(finfo.eps),
            smallest_normal=float(finfo.smallest_normal),
        )


def get_finfo(dtype: NullableCore | CoreType) -> Finfo:
    try:
        if isinstance(dtype, NullableCore):
            dtype = dtype.values
        return Finfo._from_dtype(dtype)
    except KeyError:
        raise TypeError(f"'{dtype}' is not a floating point data type.")


def get_iinfo(dtype: NullableCore | CoreType) -> Iinfo:
    try:
        if isinstance(dtype, NullableCore):
            dtype = dtype.values
        return Iinfo._from_dtype(dtype)
    except KeyError:
        raise TypeError(f"'{dtype}' is not an integer data type.")



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause
from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from spox import Tensor, argument

import ndonnx as ndx
import ndonnx._data_types as dtypes
import ndonnx.additional as nda
from ndonnx._corearray import _CoreArray

from ._shapeimpl import UniformShapeOperations
from ._utils import validate_core

if TYPE_CHECKING:
    from ndonnx._array import Array
    from ndonnx._data_types import Dtype


class CoreOperationsImpl(UniformShapeOperations):
    def make_array(
        self,
        shape: tuple[int | None | str, ...],
        dtype: Dtype,
        eager_value: np.ndarray | None = None,
    ) -> Array:
        if not isinstance(dtype, dtypes.CoreType):
            return NotImplemented
        return ndx.Array._from_fields(
            dtype,
            data=_CoreArray(
                dtype._parse_input(eager_value)["data"]
                if eager_value is not None
                else argument(Tensor(dtype.to_numpy_dtype(), shape))
            ),
        )

    @validate_core
    def make_nullable(self, x: Array, null: Array) -> Array:
        if null.dtype != ndx.bool:
            raise TypeError("'null' must be a boolean array")

        return ndx.Array._from_fields(
            dtypes.into_nullable(x.dtype),
            values=x.copy(),
            null=ndx.broadcast_to(null, nda.shape(x)),
        )

    @validate_core
    def where(self, condition, x, y):
        if x.dtype != y.dtype:
            target_dtype = ndx.result_type(x, y)
            x = ndx.astype(x, target_dtype)
            y = ndx.astype(y, target_dtype)
        return super().where(condition, x, y)



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause
from __future__ import annotations

from typing import TYPE_CHECKING, Union

import numpy as np

import ndonnx as ndx

from ._shapeimpl import UniformShapeOperations
from ._utils import assemble_output_recurse, validate_core

if TYPE_CHECKING:
    from ndonnx._array import Array
    from ndonnx._data_types import CoreType, StructType

    Dtype = Union[CoreType, StructType]


class NullableOperationsImpl(UniformShapeOperations):
    @validate_core
    def fill_null(self, x: Array, value) -> Array:
        value = ndx.asarray(value)
        if value.dtype != x.values.dtype:
            value = value.astype(x.values.dtype)
        return ndx.where(x.null, value, x.values)

    @validate_core
    def make_nullable(self, x: Array, null: Array) -> Array:
        return NotImplemented

    @validate_core
    def where(self, condition, x, y):
        if x.dtype != y.dtype:
            target_dtype = ndx.result_type(x, y)
            x = ndx.astype(x, target_dtype)
            y = ndx.astype(y, target_dtype)
        return super().where(condition, x, y)

    def make_array(
        self,
        shape: tuple[int | None | str, ...],
        dtype: Dtype,
        eager_value: np.ndarray | None = None,
    ) -> Array:
        if not isinstance(dtype, ndx.Nullable):
            return NotImplemented

        if eager_value is not None:
            eager_values = dtype._parse_input(eager_value)
            values_np = assemble_output_recurse(dtype.values, eager_values["values"])
            values = ndx.asarray(values_np, dtype=dtype.values)
            null = ndx.asarray(
                np.broadcast_to(
                    assemble_output_recurse(dtype.null, eager_values["null"]),
                    values_np.shape,
                ),
                dtype=dtype.null,
            )
        else:
            values = ndx.array(shape=shape, dtype=dtype.values)
            null = ndx.array(shape=shape, dtype=dtype.null)

        return ndx.Array._from_fields(
            dtype,
            values=values,
            null=null,
        )



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import functools
import operator
from typing import TYPE_CHECKING

import numpy as np

import ndonnx as ndx
import ndonnx._data_types as dtypes
import ndonnx._opset_extensions as opx

from ._coreimpl import CoreOperationsImpl
from ._interface import OperationsBlock
from ._nullableimpl import NullableOperationsImpl
from ._utils import binary_op, unary_op, validate_core

if TYPE_CHECKING:
    from ndonnx import Array


class _BooleanOperationsImpl(OperationsBlock):
    @validate_core
    def equal(self, x, y) -> Array:
        return binary_op(x, y, opx.equal)

    @validate_core
    def not_equal(self, x, y) -> ndx.Array:
        return self.logical_not(self.equal(x, y))

    @validate_core
    def bitwise_and(self, x, y) -> Array:
        return self.logical_and(x, y)

    @validate_core
    def bitwise_invert(self, x) -> Array:
        return self.logical_not(x)

    @validate_core
    def bitwise_or(self, x, y) -> Array:
        return self.logical_or(x, y)

    @validate_core
    def bitwise_xor(self, x, y) -> Array:
        return self.logical_xor(x, y)

    @validate_core
    def logical_and(self, x, y):
        # If one of the operands is True and the broadcasted shape can be guaranteed to be the other array's shape,
        # we can return this other array directly.
        x, y = map(ndx.asarray, (x, y))
        y_np = y.to_numpy()
        x_np = x.to_numpy()
        if x_np is not None and x_np.size == 1 and x.ndim <= y.ndim and x_np.item():
            return y.copy()
        elif y_np is not None and y_np.size == 1 and y.ndim <= x.ndim and y_np.item():
            return x.copy()
        return binary_op(x, y, opx.and_)

    @validate_core
    def logical_not(self, x) -> Array:
        return unary_op(x, opx.not_)

    @validate_core
    def logical_or(self, x, y):
        # If one of the operands is False and the broadcasted shape can be guaranteed to be the other array's shape,
        # we can return this other array directly.
        x, y = map(ndx.asarray, (x, y))
        x_np = x.to_numpy()
        y_np = y.to_numpy()
        if x_np is not None and x_np.size == 1 and x.ndim <= y.ndim and not x_np.item():
            return y.copy()
        elif (
            y_np is not None and y_np.size == 1 and y.ndim <= x.ndim and not y_np.item()
        ):
            return x.copy()
        return binary_op(x, y, opx.or_)

    @validate_core
    def logical_xor(self, x, y):
        x, y = map(ndx.asarray, (x, y))
        return binary_op(x, y, opx.xor)

    @validate_core
    def argmax(self, x, axis=None, keepdims=False) -> Array:
        return ndx.argmax(x.astype(ndx.int32), axis=axis, keepdims=keepdims)

    @validate_core
    def argmin(self, x, axis=None, keepdims=False) -> Array:
        return ndx.argmin(x.astype(ndx.int32), axis=axis, keepdims=keepdims)

    @validate_core
    def can_cast(self, from_, to) -> bool:
        if isinstance(from_, ndx.CoreType) and isinstance(to, ndx.CoreType):
            return np.can_cast(from_.to_numpy_dtype(), to.to_numpy_dtype())
        return NotImplemented

    @validate_core
    def all(self, x, *, axis=None, keepdims: bool = False):
        if isinstance(x.dtype, dtypes.NullableCore):
            x = ndx.where(x.null, True, x.values)
        if functools.reduce(operator.mul, x._static_shape, 1) == 0:
            return ndx.asarray(True, dtype=ndx.bool)
        return ndx.min(x.astype(ndx.int8), axis=axis, keepdims=keepdims).astype(
            ndx.bool
        )

    @validate_core
    def any(self, x, *, axis=None, keepdims: bool = False):
        if isinstance(x.dtype, dtypes.NullableCore):
            x = ndx.where(x.null, False, x.values)
        if functools.reduce(operator.mul, x._static_shape, 1) == 0:
            return ndx.asarray(False, dtype=ndx.bool)
        return ndx.max(x.astype(ndx.int8), axis=axis, keepdims=keepdims).astype(
            ndx.bool
        )

    @validate_core
    def ones(
        self,
        shape,
        dtype: dtypes.CoreType | dtypes.StructType | None = None,
        device=None,
    ):
        dtype = dtypes.float64 if dtype is None else dtype
        return ndx.full(shape, True, dtype=dtype)

    @validate_core
    def ones_like(
        self, x, dtype: dtypes.StructType | dtypes.CoreType | None = None, device=None
    ):
        return ndx.full_like(x, True, dtype=dtype)

    @validate_core
    def zeros(
        self,
        shape,
        dtype: dtypes.CoreType | dtypes.StructType | None = None,
        device=None,
    ):
        dtype = dtypes.float64 if dtype is None else dtype
        return ndx.full(shape, False, dtype=dtype)

    @validate_core
    def zeros_like(
        self, x, dtype: dtypes.CoreType | dtypes.StructType | None = None, device=None
    ):
        return ndx.full_like(x, False, dtype=dtype)

    @validate_core
    def empty(self, shape, dtype=None, device=None) -> ndx.Array:
        shape = ndx.asarray(shape, dtype=dtypes.int64)
        return ndx.full(shape, False, dtype=dtype)

    @validate_core
    def empty_like(self, x, dtype=None, device=None) -> ndx.Array:
        return ndx.full_like(x, False, dtype=dtype)

    @validate_core
    def nonzero(self, x) -> tuple[Array, ...]:
        return ndx.nonzero(x.astype(ndx.int8))


class BooleanOperationsImpl(CoreOperationsImpl, _BooleanOperationsImpl): ...


class NullableBooleanOperationsImpl(NullableOperationsImpl, _BooleanOperationsImpl): ...



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

import pytest

import ndonnx as ndx


@pytest.mark.parametrize(
    "dtype",
    [
        ndx.int8,
        ndx.int16,
        ndx.int32,
        ndx.int64,
        ndx.float32,
        ndx.float64,
        ndx.utf8,
        ndx.bool,
        ndx.uint8,
        ndx.uint16,
        ndx.uint32,
        ndx.uint64,
    ],
)
def test_input_output_name_backwards_compatibility(dtype):
    a = ndx.array(shape=("N",), dtype=dtype)
    model_proto = ndx.build({"input": a}, {"output": a})
    assert [node.name for node in model_proto.graph.input] == ["input"]
    assert [node.name for node in model_proto.graph.output] == ["output"]
    a = ndx.array(shape=("N",), dtype=ndx._data_types.into_nullable(dtype))
    model_proto = ndx.build({"input": a}, {"output": a})
    assert [node.name for node in model_proto.graph.input] == [
        "input_values",
        "input_null",
    ]
    assert [node.name for node in model_proto.graph.output] == [
        "output_values",
        "output_null",
    ]
